#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unistd.h>
#include <pthread.h>
#include <errno.h>
#include <ntrace.h>
#include <math.h>
#include <sys/ipc.h>
#include <sys/sem.h>
#include <sys/shm.h>
#include <signal.h>
#include <time.h>

static void * heap_thread (void * ptr);
 
typedef struct {
   char * name;
   int    count;
   double delta;
   double angle;
   double value;
} control_t;
 
control_t data[2] = { { "sin", 0, M_PI/360.0, 0.0, 0.0 },
                      { "cos", 0, M_PI/360.0, 0.0, 0.0 } };
enum { run, hold } state;
int rate = 50000000;
int sema;
int StopFlg = 0;

static
void *
GetSharedMem (key_t id)
{
   int    sid, flg;
   void   *ptr;
   struct shmid_ds shm_buf;

   flg = IPC_CREAT | SHM_R | SHM_W;
   sid = shmget (id, 0x1000, flg);
   if (sid == (-1)) return NULL;

   ptr = (void*)shmat(sid, NULL, SHM_RND);
   if (ptr == NULL) return NULL;

   if (shmctl(sid, IPC_STAT, &shm_buf) < 0) return NULL;
   else
   {
      shm_buf.shm_perm.mode = 0666;
      if (shmctl(sid, IPC_SET, &shm_buf) < 0) return NULL;
   }

   return ptr;
}

extern
double
FunctionCall(void)
{
   return data[0].value + data[1].value;
}
 
static
void *
sine_thread (void * ptr)
{
   control_t * data = (control_t *)ptr;
   struct sembuf wait = {0, -1, 0};
   float  *shm_ptr;
 
   trace_open_thread (data->name);
   shm_ptr = GetSharedMem((key_t)0x1234);

   for (;StopFlg == 0;) {
      semop(sema, &wait, 1);
      data->count++;
      data->angle += data->delta;
      data->value = sin(data->angle);

      if (shm_ptr != NULL) *shm_ptr = data->value;
   }

   pthread_exit(NULL);
}
 
static
void *
cosine_thread (void * ptr)
{
   control_t * data = (control_t *)ptr;
   struct sembuf wait = {0, -1, 0};
   float  *shm_ptr;
   struct timespec now;
 
   trace_open_thread (data->name);
   shm_ptr = GetSharedMem((key_t)0x5678);
 
   for (;StopFlg == 0;) {
      semop(sema, &wait, 1);
      data->count++;
      data->angle += data->delta;
      data->value = cos(data->angle);

      clock_gettime(CLOCK_REALTIME, &now);
      if (shm_ptr != NULL &&
          ((now.tv_nsec % 10) == (data->count % 10)))
      {
         *shm_ptr = data->value;
      }
   }

   pthread_exit(NULL);
}

static void sigint_handler(int signo)
{
   StopFlg = 1;
}

int
main (int argc, char * argv[])
{
   pthread_t thread;
   pthread_attr_t attr;
   struct sembuf trigger = {0, 2, 0};
   struct sigaction sigact;

   sigemptyset(&sigact.sa_mask);
   sigaddset(&sigact.sa_mask, SIGINT);
   sigact.sa_flags = SA_RESTART;
   sigact.sa_handler = sigint_handler;
   sigaction(SIGINT, &sigact, NULL);

   trace_begin ("/tmp/data",NULL);
   trace_open_thread ("main");

   sema = semget (IPC_PRIVATE, 1, IPC_CREAT+0666);

   pthread_attr_init(&attr);
   Pthread_create (&thread, &attr, sine_thread, &data[0]);
   
   pthread_attr_init(&attr);
   Pthread_create (&thread, &attr, cosine_thread, &data[1]);

   pthread_attr_init(&attr);
   Pthread_create (&thread, &attr, heap_thread, NULL);

   for (;StopFlg == 0;) {
      struct timespec delay = { 0, rate } ;
      nanosleep(&delay,NULL);
      if (state != hold) semop(sema,&trigger,1);
   }

   trace_end () ;
}
 
void * ptrs[5];
 
static
void *
heap_thread (void * unused)
{
   int i;
   int scenario = -1;
   void * ptr;
   void * ptr2;
   int * * iptr;
 
   extern void * alloc_ptr (int size, int swtch);
   extern void free_ptr (void * ptr, int swtch);
 
   for (;StopFlg == 0;) {
      sleep (3);
      switch (scenario) {
      case 1:
         // Use of freed pointer
         ptr = alloc_ptr(1024,3);
         free_ptr(ptr,2);
         memset (ptr, 47, 64);
         break;
      case 2:
         // Double-free
         ptr = alloc_ptr(1024,3);
         free_ptr(ptr,2);
         free(ptr);
         break;
      case 3:
         // Overwriting past end of an allocated block
#define MyString "mystring"
         ptr  = alloc_ptr(strlen(MyString),  2);
         ptr2 = alloc_ptr(strlen(MyString)+1,2);
         strcpy (ptr, MyString); // oops -- forgot the zero-byte
         strcpy (ptr2,MyString);
         break;
      case 4:
         // Uninitialized use
         iptr = (int * *) alloc_ptr(sizeof(void*),2);
         if (*iptr) **iptr = 2778;
         break;
      case 5:
         // Leak -- all references to block removed
         ptr = alloc_ptr(37,1);
         ptr = 0;
         break;
      case 6:
         // Some more allocations we'll check on...
         ptrs[0] = alloc_ptr(1024*1024,3);
         ptrs[1] = alloc_ptr(1024,2);
         ptrs[2] = alloc_ptr(62,1);
         ptrs[3] = alloc_ptr(4564,3);
         ptrs[4] = alloc_ptr(8177,3);
         break;
      }
 
      (void) malloc(1);
      scenario = 0;
   }

   pthread_exit(NULL);
}
 
void * func3 (int size, int count)
{
   return malloc(size);
}
 
void * func2 (int size, int count)
{
   if (--count > 0) return func3(size,count);
   return malloc(size);
}
 
void * func1 (int size, int count)
{
   if (--count > 0) return func2(size,count);
   return malloc(size);
}
 
void free3 (void * ptr, int count)
{
   free(ptr);
}
 
void free2 (void * ptr, int count)
{
   if (--count > 0) {
      free3(ptr,count);
      return;
   }
   free(ptr);
}
 
void free1 (void * ptr, int count)
{
   if (--count > 0) {
      free2(ptr,count);
      return;
   }
   free(ptr);
}
 
void * alloc_ptr (int size, int count)
{
   return func1(size,count);
}
 
void free_ptr (void * ptr, int count)
{
   free1(ptr,count);
}

