#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sched.h>
#include <sys/stat.h>
#include <fbsched.h>
#include <errno.h>
#define NUM_PROCS 4
#define START 1
#define STOP 0
#define PROGRAM1 "./prog1"
#define PROGRAM2 "./prog2"
#define IDLE0 "idle0"
#define IDLE1 "idle1"
#define CLOCK1 "/dev/rcim/rtc0"

void
cleanup(int fbs_id)
{
	fbsrunrtc(fbs_id, 0);	/* stop timing device */
	fbsdetach(fbs_id);	/* detach timing device */
	fbsremove(fbs_id, -1);	/* remove frequency-based scheduler */
}

int main()
{
	struct fbsconfig_ds fbs_buf;
	struct pgm2_ds sched_buf;
	struct pmqry_ds pm_buf[NUM_PROCS];
	struct fbsinfo_ds info_buf;
	struct fbscycle_ds cycle_buf;
	int idle0_fpid;	/* fpid for idle on cpu 0 */
	int idle1_fpid;	/* fpid for idle on cpu 1 */
	int pgm1_fpid;	/* fpid for testprogram 1 */
	int pgm2_fpid;	/* fpid for testprogram 2 */
	int cpu;
	int istat;
	int pmflg;
	int intrflg;
	int i;
	int count;
	int resolution;
	char name[1024];	/* programa??s full or orelative path name */
	char *pgmname[4];	/* programa??s name */
	FILE *fp;
	/* Open file to store performance information */

	fp= fopen("pmresults", "w+");
	if (fp == NULL)
		printf("open failed errno = %d\n", errno);
	/*
	* CONFIGURE SCHEDULER
	*/
	fbs_buf.key = 37;	/* scheduler key */
	fbs_buf.cycles = 10;	/* number of cycles per frame */
	fbs_buf.progs = 5;	/* max. number of programs per cycle */
	fbs_buf.max = 10;	/* max. number of programs allowed on the fbs */
	fbs_buf.reset = -1;	/* kill/remove processes currently scheduled */

	/* owner/group read/write */
	fbs_buf.configflg = 0664;
	istat = fbsconfigure(&fbs_buf);
	if (istat != 0) {
		printf("could not configure scheduler: errno = %d\n" ,errno);
		exit(1);
	}
	/*
	* SCHEDULE test program1 PROGRAM1
	*/
	sched_buf.name_ptr = PROGRAM1;
	sched_buf.cid = SCHED_FIFO; /* first-in-first out (FIFO) policy */
	sched_buf.prior = 19;
	sched_buf.param = 0;	/* optional initiation parameter */
	sched_buf.period = 2;	/* time between wakeups */
	sched_buf.cycle = 0;	/* starting base cycle */
	sched_buf.halt = 0;	/* halt on overrun */

	/* Set cpu mask to schedule testprogram1 on cpu 0 */
	sched_buf.cpu = 1;
	istat = sched_pgmadd(fbs_buf.fbs_id, &sched_buf);
	printf("fbsid = %d\n", fbs_buf.fbs_id);
	if (istat != 0) {
		printf("could not schedule %s on cpu %d : errno = %d\n",
		sched_buf.name_ptr, sched_buf.cpu>>1, errno);
		cleanup(fbs_buf.fbs_id);
		return(1);
	}

	pgm1_fpid = sched_buf.fpid;
	printf("pgm1 fpid = %d\n", pgm1_fpid);
	/*
	* SCHEDULE test program2 PROGRAM2
	*/
	sched_buf.name_ptr = PROGRAM2;
	sched_buf.prior = 19;
	sched_buf.cid = SCHED_FIFO; /* first-in-first out (FIFO) policy */
	sched_buf.param = 0;	/* optional initiation parameter */
	sched_buf.period = 2;	/* time between wakeups */
	sched_buf.cycle = 1;	/* starting base cycle */
	sched_buf.halt = 0;	/* halt on overrun */
	/* Set cpu mask to schedule testprogram2 on cpu 1 */
	sched_buf.cpu = 2;
	istat = sched_pgmadd(fbs_buf.fbs_id, &sched_buf);
	printf("fbsid = %d\n", fbs_buf.fbs_id);
	if (istat != 0) {
		printf("could not schedule %s on cpu %d : errno = %d\n",
		sched_buf.name_ptr, sched_buf.cpu>>1, errno);
		cleanup(fbs_buf.fbs_id);
		return(1);
	}
	pgm2_fpid = sched_buf.fpid;
	printf("pgm2 fpid = %d\n", pgm2_fpid);

	/*
	* SCHEDULE IDLE ON CPU 0
	*/
	sched_buf.name_ptr = "/idle";
	sched_buf.cid = SCHED_OTHER; /* first-in-first out (FIFO) policy */
	sched_buf.prior = 0;
	sched_buf.param = 0;	/* optional initiation parameter */
	sched_buf.period = 1;	/* time between wakeups */
	sched_buf.cycle = 0;	/* starting base cycle */
	sched_buf.halt = 0;	/* halt on overrun */
	/* Set cpu mask to schedule idle on cpu 0 */
	sched_buf.cpu = 1;	/* mask = 1<<cpu */
	istat = sched_pgmadd(fbs_buf.fbs_id, &sched_buf);
	printf("fbsid = %d\n", fbs_buf.fbs_id);
	if (istat != 0) {
		printf("could not schedule %s on cpu %d : errno = %d\n",
		sched_buf.name_ptr, sched_buf.cpu, errno);
		cleanup(fbs_buf.fbs_id);
		return(1);
	}
	idle0_fpid = sched_buf.fpid;
	printf("idle0 fpid = %d\n", idle0_fpid);
	
	/*
	* SCHEDULE IDLE ON CPU 1
	*
	* The only parameter required for /idle is the CPU.
	*/
	sched_buf.name_ptr = "/idle";
	sched_buf.prior = 0;
	sched_buf.cid = SCHED_OTHER; /* first-in-first out (FIFO) policy */
	sched_buf.param = 0;	/* optional initiation parameter */
	sched_buf.period = 1;	/* time between wakeups */
	sched_buf.cycle = 0;	/* starting base cycle */
	sched_buf.halt = 0;	/* halt on overrun */
	/* Set cpu mask to schedule idle on cpu 1 */
	sched_buf.cpu = 2;	/* mask = 1<<cpu */
	istat = sched_pgmadd(fbs_buf.fbs_id, &sched_buf);
	if (istat != 0) {
		printf("could not schedule %s on cpu %d : errno = %d\n",
		sched_buf.name_ptr, sched_buf.cpu, errno);
		cleanup(fbs_buf.fbs_id);
		return(1);
	}
	idle1_fpid = sched_buf.fpid;
	printf("idle1 fpid = %d\n", idle1_fpid);

	/*
	* ATTACH/SET REAL-TIME CLOCK
	* Set the clock to interrupt every 10 msecs.
	*/
	count = 10000;
	resolution = 1;
	istat = fbsattach(fbs_buf.fbs_id, CLOCK1);
	if (istat != 0) {
		printf("could not attach timing source: errno = %d\n", errno);
		cleanup(fbs_buf.fbs_id);
		return(1);
	}
	istat = fbssetrtc(fbs_buf.fbs_id, count, resolution);
	if (istat != 0) {
		printf("could not set rtc: errno = %d\n", errno);
		cleanup(fbs_buf.fbs_id);
		return(1);
	}
	istat = fbsrunrtc(fbs_buf.fbs_id, START);
	if (istat != 0) {
		printf("could not start rtc: errno = %d\n", errno);
		cleanup(fbs_buf.fbs_id);
		return(1);
	}

	/*
	* START PERFORMANCE MONITORING
	*/
	pmflg = 1;
	/* zero out the "name" variable (fpid must be specified).
	* Ulimately, the "name" variable can be used to store
	* the full or relative path name of a test program.
	*/
	bzero(name, sizeof(name));
	cpu = 0; /* not used if fpid is being used */
	/* start performance monitoring for testprogram1 */
	istat = pmprogram(fbs_buf.fbs_id, name, cpu, pgm1_fpid, pmflg);
	if (istat != 0) {
		printf("could not start pm for testprogram1 : errno = %d\n", errno);
		cleanup(fbs_buf.fbs_id);
		return(1);
	}

	/* start performance monitoring for testprogram2 */
	istat = pmprogram(fbs_buf.fbs_id, name, cpu, pgm2_fpid, pmflg);
	if (istat != 0) {
		printf("coul dnot start pm for testprogram2 : errno = %d\n", errno);
		cleanup(fbs_buf.fbs_id);
		return(1);
	}

	/* start performance monitoring for idle on cpu 0 */
	istat = pmprogram(fbs_buf.fbs_id, name, cpu, idle0_fpid, pmflg);
		if (istat != 0) {
		printf("could not start pm for idle0 : errno = %d\n", errno);
		cleanup(fbs_buf.fbs_id);
		return(1);
	}

	/* start performance monitoring for idle on cpu 1 */
	istat = pmprogram(fbs_buf.fbs_id, name, cpu, idle1_fpid, pmflg);
	if (istat != 0) {
		printf("could not start pm for idle1 : errno = %d\n", errno);
		cleanup(fbs_buf.fbs_id);
		return(1);
	}

	/*
	* START SCHEDULING
	*/
	intrflg = 1;
	istat = fbsintrpt(fbs_buf.fbs_id, intrflg);
	if (istat != 0) {
		printf("could not start scheduler : errno = %d\n", errno);
		cleanup(fbs_buf.fbs_id);
		return(1);
	}

	/*
	* QUERY PERFORMANCE MONITOR VALUES
	* 1 second = 100 cycles 1 minute = 600 frames
	* Query once per second for 1 minute
	*/
	pm_buf[0].fpid = pgm1_fpid; pgmname[0] = PROGRAM1;
	pm_buf[1].fpid = pgm2_fpid; pgmname[1] = PROGRAM2;
	pm_buf[2].fpid = idle0_fpid; pgmname[2] = IDLE0;
	pm_buf[3].fpid = idle1_fpid; pgmname[3] = IDLE1;

	sleep(1); /* sleep for a while */
	printf("Please wait, performance information being gathered. \n");

	cycle_buf.cframe = 0;
	while ((cycle_buf.cframe < 600)) {
		istat = fbscycle(fbs_buf.fbs_id, &cycle_buf);
		if (istat != 0) {
			printf("could not query fbscycle: errno = %d\n", errno);
			cleanup(fbs_buf.fbs_id);
			return(1);
		}
		istat = pmqrylist(fbs_buf.fbs_id, pm_buf, NUM_PROCS);
		if (istat != 0) {
			printf("could not query process: errno = %d\n", errno);
			cleanup(fbs_buf.fbs_id);
			return(1);
		}

		fprintf(fp,"*****************************************\n");
		fprintf(fp,"\tFRAME=%d, CYCLE=%d\n", cycle_buf.cframe,
		cycle_buf.ccycle);
		fprintf(fp,"*****************************************\n\n");
		/* Write performance information for each
		* process into a file
		*/
		for (i = 0; i < NUM_PROCS; i++) {
			fprintf(fp,"pgm=%s, fpid=%d\n", pgmname[i],
			pm_buf[i].fpid);
			fprintf(fp,"--------------------------------------\n");
			fprintf(fp,"last cycle secs:nsecs = %d:%d\n",
			pm_buf[i].lastcyc_tm.tv_sec,
			pm_buf[i].lastcyc_tm.tv_nsec);
			fprintf(fp,"total cycles = %d\n",pm_buf[i].tot_cycles);
			fprintf(fp, "total cycles secs:nsecs = %d:%d\n",
			pm_buf[i].tot_cycles_tm.tv_sec,
			pm_buf[i].tot_cycles_tm.tv_nsec);
			fprintf(fp, "overruns = %d\n",pm_buf[i].overruns);
			fprintf(fp, "mincyc secs:nsecs = %d:%d\n",
			pm_buf[i].mincyc_tm.tv_sec,
			pm_buf[i].mincyc_tm.tv_nsec);
			fprintf(fp, "mincyc_cycle = %d\n",
			pm_buf[i].mincyc_cycle);
			fprintf(fp, "mincyc_frame = %d\n",
			pm_buf[i].mincyc_frame);
			fprintf(fp, "maxcyc secs:nsecs = %d:%d\n",
			pm_buf[i].maxcyc_tm.tv_sec,
			pm_buf[i].maxcyc_tm.tv_nsec);
			fprintf(fp, "maxcyc_cycle = %d\n",
			pm_buf[i].maxcyc_cycle);
			fprintf(fp, "maxcyc_frame = %d\n",
			pm_buf[i].maxcyc_frame);
			fprintf(fp, "minframe secs:nsecs = %d:%d\n",
			pm_buf[i].minframe_tm.tv_sec,
			pm_buf[i].minframe_tm.tv_nsec);
			fprintf(fp, "minframe = %d\n", pm_buf[i].minframe);
			fprintf(fp, "maxframe secs:nsecs = %d:%d\n",
			pm_buf[i].maxframe_tm.tv_sec,
			pm_buf[i].maxframe_tm.tv_nsec);
			fprintf(fp, "maxframe = %d\n", pm_buf[i].maxframe);
			fprintf(fp, "\n");
		}
		sleep(1); /* sleep for a while */
	}

	printf("Performance data has been gathered. \n");
	if (istat != 0)
		printf("istat = %d\n", istat);
	/* Stop PM on CPUs 0 and 1 */
	pmmonitor(fbs_buf.fbs_id, 0, 3);
	/* Stop the clock */
	istat = fbsrunrtc(fbs_buf.fbs_id, STOP);
	if (istat != 0) {
		printf("could not stop rtc: errno = %d\n", errno);
		cleanup(fbs_buf.fbs_id);
		return(1);
	}

	/* Detach from the fbs */
	istat = fbsdetach(fbs_buf.fbs_id);
	if (istat != 0) {
		printf("could not dettach timing source: errno = %d\n", errno);
		cleanup(fbs_buf.fbs_id);
		return(1);
	}
	/* Remove the fbs */
	istat = fbsremove(fbs_buf.fbs_id, -1);
	if (istat != 0) {
		printf("could not remove timing source: errno = %d\n", errno);
		cleanup(fbs_buf.fbs_id);
		return(1);
	}
	cleanup(fbs_buf.fbs_id);

	return(0);
}

