/*
	POSIX RedHawk BenchMark test program version 3.0
*/
#include <stdio.h>
#include <time.h>
#include <signal.h>
#include <errno.h>
#include <sched.h>
#include <sys/types.h>
#include <unistd.h>
#include <mpadvise.h>
#include <cpuset.h>
#include <sys/mman.h>
#include <stdlib.h>
#include <math.h>


#define MAXJITTER (30)
#define BOOTMAXJITTER (100)

extern int errno;
static BOOT_TEST=0;
static FILE *logfd;

//#define HANDLER

void
readtime(struct timespec *nowtime)
{
        clock_gettime(CLOCK_REALTIME,nowtime);
}

void
adjust(struct timespec *start,struct timespec *finish, float *realtime)
{
        int     sec,nsec;

        sec = finish->tv_sec - start->tv_sec;
        nsec = finish->tv_nsec - start->tv_nsec;
        if (nsec<0)
        {
                sec --;
                nsec += 1000000000;
        }
        *realtime = (float)(nsec/1000)+(float)(sec*1000000);
}

timer_t
create_posix_timer(int signum,int sec,int nsec)
{
        static struct sigevent event;
        static timer_t timer;
        static struct itimerspec itspec;

        /* Create the POSIX timer to generate signum */
        event.sigev_notify = SIGEV_SIGNAL;
        event.sigev_signo = signum;
        if (timer_create((clockid_t)CLOCK_REALTIME,&event,&timer)==(-1))
        {
                fprintf(stderr, "create_posix_timer() Cannot create timer - %s\n",
                strerror(errno));
                return ((timer_t)-1);
        }
        /*      Set the initial delay and period of the timer.
        This also arms the timer */
	clock_gettime(CLOCK_REALTIME,&itspec.it_value);
        itspec.it_value.tv_sec += 3;
        itspec.it_interval.tv_sec = sec;
        itspec.it_interval.tv_nsec =  nsec;
        if (timer_settime( timer, TIMER_ABSTIME, &itspec, NULL)==(-1))
        {
                return ((timer_t)-1);
        }
        return(timer);
}


#ifdef HANDLER
int loop_flag =1;
void interval_hadler(int no, siginfo_t * info , void * p)
{
        printf("interval_hadler(%d)\n",no);
}
#endif

#define MSEC 10
#define MAXCOUNT 1001
void timer_test(float jitter)
{
        int signum,status;
        static struct sigaction newact;
        static sigset_t set,oset;
#ifdef HANDLER
        static sigset_t unset;
#endif
        static timer_t timer_id;
	static siginfo_t info;
        static struct timespec t0[MAXCOUNT];
	static float real[MAXCOUNT],min,max,avr;
	static int count = 0;

printf("********************************************* \n");
printf("****  posix timer test wait for %d sec **** \n",MSEC);
printf("********************************************* \n");
        if (sigprocmask(0,NULL,&set)==(-1))
        {
                fprintf(stderr, "Cannot get sigprocmask - %s\n",strerror(errno));
                exit(1);
        }
        sigaddset(&set,SIGRTMIN);
        if (sigprocmask(SIG_SETMASK,&set,&oset)==(-1))
        {
                fprintf(stderr, "%s: Cannot set sigprocmask - %s\n", strerror(errno));
                exit(1);
        }

        signum = SIGRTMIN;
        timer_id = create_posix_timer(signum,0,MSEC * 1000000);
        if (timer_id<0)
        {
                fprintf(stderr, "%s: Cannot set posix timer - %s\n", strerror(errno));
                exit(1);
	}
	count =0;
#ifdef HANDLER
        sigemptyset(&newact.sa_mask);
        sigaddset(&newact.sa_mask,signum);
        newact.sa_sigaction = interval_hadler;
        newact.sa_flags = SA_SIGINFO|SA_RESTART;
        if (sigaction(signum, &newact, 0)==(-1))
        {
                fprintf(stderr, "%s: Cannot set sigaction - %s\n", strerror(errno));
                exit(1);
        }
        sigfillset(&unset);
        sigdelset(&unset,SIGRTMIN);
#endif
        do
        {
#ifdef HANDLER
                status = sigsuspend(&unset);
#else
                status = sigwaitinfo(&set,&info);
#endif
        	readtime(&t0[count]);
		count++;
        } while (count<MAXCOUNT);
	timer_delete(timer_id);
	min=1000000000;
	avr =max=0;
        for(count=1;count<MAXCOUNT;count++)
	{
                adjust(&t0[count-1],&t0[count],&real[count]);
		avr += real[count];
		if (min> real[count]) min =  real[count];
		if (max< real[count]) max =  real[count];
		if (logfd) fprintf(logfd,"%f\n",real[count]);
	}
	avr /= (float)(MAXCOUNT-1);
	printf(" min %f avr %f max %f\t",min,avr,max);
	if (fabs(max-min)<=jitter)
	{
		printf("GOOD(fabs(%f)<%f micro sec)\n",(max-min),jitter);
	}
	else
	{
		printf("FAIL(fabs(%f)>%f micro sec)\n",(max-min),jitter);
	}
}

void counter_test(float jitter)
{
        struct timespec t0,t1;
        struct timespec w0={0,0},w1={0,0};
        float real,over;
        int i,usec,nsec,res,ovr,old;


printf("*********************************************** \n");
printf("****  posix counter test                   **** \n");
printf("*********************************************** \n");
        clock_getres(CLOCK_REALTIME,&w0);
	res = w0.tv_nsec;
        printf("clock resolution is %d nsec\n",res);
	ovr = 0;
	for(i=0;i<100;i++)
	{
        	w0.tv_sec=0;
        	w0.tv_nsec= 10*1000*1000;
        	readtime(&t0);
        	nanosleep(&w0,&w1);
        	readtime(&t1);
        	adjust(&t0,&t1,&over);
		ovr += (((int)over * 1000) - w0.tv_nsec);
	}
	ovr /= 100;
        printf("nanosleep resolution test overhead is %d nsec\n",ovr);
	fflush(stdout);
        for(usec=10;usec<1000000;usec=usec<<1)
        {
                w0.tv_sec=0;
                w0.tv_nsec=1000 * usec - ovr;
		if (w0.tv_nsec<0) w0.tv_nsec = 0;
                readtime(&t0);
                nanosleep(&w0,&w1);
                readtime(&t1);
                adjust(&t0,&t1,&real);
                printf("request %10d usec real %10.3f usec ret %d nsec\t",
			usec,real,w1.tv_nsec);
		if (fabs(real-usec)<=jitter)
		{
			printf("GOOD(fabs(%f)<%f micro sec)\n",(real-usec),jitter);
		}
		else
		{
			printf("FAIL(fabs(%f)>%f micro sec)\n",(real-usec),jitter);
		}
        }
}

int
main(int argc, char **argv)
{
	static struct sched_param myparam;
	cpuset_t* setp;
	int ret;
        extern int errno;
        extern int optind;
        extern char *optarg;
        register int c;
        char    *progname;

        progname = argv[0];
        while((c = getopt(argc, argv, "B1:")) != EOF){
                switch(c) {
                        case 'B':
                                BOOT_TEST=1;
                                break;
                        case 'l':
                                logfd=fopen(optarg,"w");
                                break;
                        default:
                                fprintf(stderr,
                                "usage: %s [-B] B:BOOT_CPU_TEST\n",
                                progname);
                                exit(1);
                }
        }
#if 0
        if (optind != argc) {
                for (c = optind; c < argc; c++) {
                        if (access(argv[c], 4)) {
                                fprintf(stderr,
                                "%s: \"%s\" %s\n",
                                 progname,argv[c],strerror(errno));
                                exit(1);
                        }
                }
        }
#endif


	mlockall(MCL_CURRENT|MCL_FUTURE);

	setp = cpuset_alloc();
        myparam.sched_priority = sched_get_priority_max(SCHED_FIFO);
        sched_setscheduler(getpid(),SCHED_FIFO,&myparam);
	system("/usr/bin/shield -r");
#if 0
	system("/usr/bin/shield -c");
	system("/usr/bin/cpu");
	timer_test(MAXJITTER);
	counter_test(MAXJITTER);
	printf("\n\n");
#endif
	system("/usr/bin/shield -p 1 -i 1");
	system("/usr/bin/shield -c");
	system("/usr/bin/cpu");
	cpuset_set_string(setp, "2");
	ret = mpadvise (MPA_PRC_SETBIAS,MPA_PID,0,setp);
	if (ret!=MPA_FAILURE)
        {
		printf("running 2nd CPU\n");
        }
	else
	{
		printf("Can not running 2nd CPU\n");
	}
	timer_test(MAXJITTER);
	counter_test(MAXJITTER);
	printf("\n\n\n");

	if (BOOT_TEST)
	{
		cpuset_set_string(setp, "1");
		ret = mpadvise (MPA_PRC_SETBIAS,MPA_PID,0,setp);
		if (ret!=MPA_FAILURE)
        	{
			printf("running boot CPU\n");
        	}
		else
		{
			printf("Can not running boot CPU\n");
		}
		timer_test(BOOTMAXJITTER);
		counter_test(BOOTMAXJITTER);
		printf("\n\n");
	}
	system("/usr/bin/shield -a 1");
	system("/usr/bin/shield -c");
	system("/usr/bin/cpu");
	cpuset_set_string(setp, "2");
	ret = mpadvise (MPA_PRC_SETBIAS,MPA_PID,0,setp);
	if (ret!=MPA_FAILURE)
        {
		printf("running 2nd CPU\n");
        }
	else
	{
		printf("Can not running 2nd CPU\n");
	}
	timer_test(MAXJITTER);
	counter_test(MAXJITTER);
	if (logfd) fclose(logfd);

}
