#include <stdio.h>
#include <errno.h>
#include <sys/types.h>
#include <time.h>
#include <signal.h>
#include <string.h>
#include <unistd.h>
#include <mpadvise.h>
#include <pthread.h>
#include <locale.h>
#include <cpuset.h>
#include <stdlib.h>
#include <math.h>

#define CYCLE		1000
#define THREAD_NO	3

char 	*progname,*lang;

volatile sig_atomic_t loop_flag=1,GlobalCounter=0;
sigset_t diset,oset,eiset;

#define TESTNUMBER	10000
#define MAXJITTER	(30)

pthread_cond_t Condition[THREAD_NO];
#define COND_SIGNAL	pthread_cond_signal
#define COND_WAIT	pthread_cond_wait
pthread_mutex_t Mutex[THREAD_NO];
#define MUTEX_LOCK	pthread_mutex_lock
#define MUTEX_UNLOCK	pthread_mutex_unlock

int Event[THREAD_NO]={0,0,0};
float SwitchTime1[TESTNUMBER/2];
float SwitchTime2[TESTNUMBER/2];
float IntervalTime[TESTNUMBER];
struct timespec StartTime={0,0},EndTime={0,0};
struct timespec CurrrentTime={0,0};


void
readtime(struct timespec *nowtime)
{
	clock_gettime(CLOCK_REALTIME,nowtime);
}

void
adjust(struct timespec *start,struct timespec *finish, float *realtime)
{
	int     sec,nsec;

	sec = finish->tv_sec - start->tv_sec;
	nsec = finish->tv_nsec - start->tv_nsec;
	if (nsec<0)
	{
		sec --;
		nsec += 1000000000;
	}
	*realtime = (float)(nsec/1000)+(float)(sec*1000000);
}


timer_t
create_posix_timer(int signum,int sec,int nsec)
{
        static struct sigevent event;
	static timer_t timer;
        static struct itimerspec itspec;

	/* Create the POSIX timer to generate signum */
        event.sigev_notify = SIGEV_SIGNAL;
        //event.sigev_notify = SIGEV_THREAD;
        event.sigev_signo = signum;
	if (timer_create((clockid_t)CLOCK_REALTIME,&event,&timer)==(-1))
        {
                fprintf(stderr, "create_posix_timer() Cannot create timer - %s\n",
                strerror(errno));
		return ((timer_t)-1);
        }
	/*      Set the initial delay and period of the timer.
        This also arms the timer */
	clock_gettime(CLOCK_REALTIME,&itspec.it_value);
        itspec.it_value.tv_sec += 1;
        itspec.it_interval.tv_sec = sec;
        itspec.it_interval.tv_nsec =  nsec;
        if (timer_settime( timer, TIMER_ABSTIME, &itspec, NULL)==(-1))
        {
		return ((timer_t)-1);
        }
	return(timer);
}


void interval_hadler(int sig)
{
#ifdef DEBUG
        printf("interval_hadler(%d)\n",GlobalCounter);
#endif
	if(Event[0]!=THREAD_NO) return;

	readtime(&StartTime);
	if(CurrrentTime.tv_sec)
	{
		adjust(&CurrrentTime,&StartTime,&IntervalTime[GlobalCounter]);
	}
	CurrrentTime.tv_sec = StartTime.tv_sec;
	CurrrentTime.tv_nsec = StartTime.tv_nsec;
	switch(GlobalCounter%10)
	{
		case 0:
			MUTEX_LOCK(&Mutex[1]); /* lock */
				Event[1]++;
				COND_SIGNAL(&Condition[1]);
			MUTEX_UNLOCK(&Mutex[1]); /* unlock */
		break;
		case 1:
			MUTEX_LOCK(&Mutex[2]); /* lock */
				Event[2]++;
				COND_SIGNAL(&Condition[2]);
			MUTEX_UNLOCK(&Mutex[2]); /* unlock */
		break;
		case 2:
			MUTEX_LOCK(&Mutex[1]); /* lock */
				Event[1]++;
				COND_SIGNAL(&Condition[1]);
			MUTEX_UNLOCK(&Mutex[1]); /* unlock */
		break;
		case 3:
			MUTEX_LOCK(&Mutex[2]); /* lock */
				Event[2]++;
				COND_SIGNAL(&Condition[2]);
			MUTEX_UNLOCK(&Mutex[2]); /* unlock */
		break;
		case 4:
			MUTEX_LOCK(&Mutex[1]); /* lock */
				Event[1]++;
				COND_SIGNAL(&Condition[1]);
			MUTEX_UNLOCK(&Mutex[1]); /* unlock */
		break;
		case 5:
			MUTEX_LOCK(&Mutex[2]); /* lock */
				Event[2]++;
				COND_SIGNAL(&Condition[2]);
			MUTEX_UNLOCK(&Mutex[2]); /* unlock */
		break;
		case 6:
			MUTEX_LOCK(&Mutex[1]); /* lock */
				Event[1]++;
				COND_SIGNAL(&Condition[1]);
			MUTEX_UNLOCK(&Mutex[1]); /* unlock */
		break;
		case 7:
			MUTEX_LOCK(&Mutex[2]); /* lock */
				Event[2]++;
				COND_SIGNAL(&Condition[2]);
			MUTEX_UNLOCK(&Mutex[2]); /* unlock */
		break;
		case 8:
			MUTEX_LOCK(&Mutex[1]); /* lock */
				Event[1]++;
				COND_SIGNAL(&Condition[1]);
			MUTEX_UNLOCK(&Mutex[1]); /* unlock */
		break;
		case 9:
			MUTEX_LOCK(&Mutex[2]); /* lock */
				Event[2]++;
				COND_SIGNAL(&Condition[2]);
			MUTEX_UNLOCK(&Mutex[2]); /* unlock */
		break;
	}
	GlobalCounter++;
}



void *thread_main0(void *arg)
{
	int class = 3;				/* FIFO Schedule */
	int priority = 1;			/* RealTime Priority min */
	struct sched_param param;	/* RealTime Priority min */
	cpuset_t	*cpumask;
	cpu_t		cpu = 1;
	volatile int thread_no;
	int i,sig;
	float min,max,avr;
	struct timespec w0 = {0,100 * 1000* 1000};
	struct timespec wait = {0,200* 1000};
	sigset_t set;

        sigemptyset(&set);
        sigaddset(&set,SIGRTMIN+2);
        sigaddset(&set,SIGINT);
	if (pthread_getschedparam(pthread_self(),&class,&param))
	{
                fprintf(stderr, "pthread %d Cannot get priority %s\n",
                pthread_self(),strerror(errno));
	}
	param.sched_priority = priority;
	if (pthread_setschedparam(pthread_self(),SCHED_FIFO,(const struct sched_param *)&param))
	{
                fprintf(stderr, "pthread %d Cannot set priority %s\n",
                pthread_self(),strerror(errno));
	}
#ifdef CCURRT
	cpumask = cpuset_alloc();
	cpuset_init(cpumask);
	cpuset_set_cpu(cpumask,cpu,1);/* 2nd CPU set */
	if ( mpadvise(MPA_PRC_SETBIAS,MPA_TID,0,cpumask)<0)
	{
                fprintf(stderr, "thread %d Cannot bind processor %s\n",
                pthread_self(),strerror(errno));
	}
        if (mpadvise (MPA_PRC_GETRUN,MPA_TID,0,cpumask)!=MPA_FAILURE)
        {
		char * cpustr;
                cpustr = cpuset_get_string(cpumask);
                printf("Thread %d is \"%s\" running\n",*(int*)arg,cpustr);
                free(cpustr);
        }
	cpuset_free(cpumask);

#endif
	MUTEX_LOCK(&Mutex[0]);
	Event[0]++;
	MUTEX_UNLOCK(&Mutex[0]);
	do
	{
		MUTEX_LOCK(&Mutex[0]);
		thread_no = Event[0];
		MUTEX_UNLOCK(&Mutex[0]);
		nanosleep(&w0,0);
	} while (thread_no<THREAD_NO);
        if (pthread_sigmask(SIG_SETMASK,&eiset,&oset)==(-1))
        {
                fprintf(stderr, "Cannot set sigprocmask - %s\n",
                strerror(errno));
	}
        do
        {
		//sigwait(&set,&sig);interval_hadler(sig);
		sigsuspend(&eiset);
		if(GlobalCounter>=TESTNUMBER) loop_flag=0;
        } while (loop_flag);
	COND_SIGNAL(&Condition[1]);
	COND_SIGNAL(&Condition[2]);
	min=99999999999.0,max=0,avr=0;
	for(i=0;i<TESTNUMBER/2;i++)
	{
		if(min>SwitchTime1[i]) min = SwitchTime1[i];
		if(max<SwitchTime1[i]) max = SwitchTime1[i];
		avr += SwitchTime1[i];
	}
	printf("Switch1 %10.3f: %10.3f : %10.3f\t",min,avr/(float)(TESTNUMBER/2),max);
        if (fabs(max-min)<=MAXJITTER)
        {
                printf("GOOD(fabs(%f)<%d micro sec)\n",max-min,MAXJITTER);
        }
        else
        {
                printf("FAIL(fabs(%f)>%d micro sec)\n",max-min,MAXJITTER);
        }

	min=99999999999.0,max=0,avr=0;
	for(i=0;i<TESTNUMBER/2;i++)
	{
		if(min>SwitchTime2[i]) min = SwitchTime2[i];
		if(max<SwitchTime2[i]) max = SwitchTime2[i];
		avr += SwitchTime2[i];
	}
	printf("Switch2 %10.3f: %10.3f : %10.3f\t",min,avr/(float)(TESTNUMBER/2),max);
        if (fabs(max-min)<=MAXJITTER)
        {
                printf("GOOD(fabs(%f)<%d micro sec)\n",max-min,MAXJITTER);
        }
        else
        {
                printf("FAIL(fabs(%f)>%d micro sec)\n",max-min,MAXJITTER);
        }



	min=99999999999.0,max=0,avr=0;
	for(i=1;i<TESTNUMBER;i++)
	{
		if(min>IntervalTime[i]) min = IntervalTime[i];
		if(max<IntervalTime[i]) max = IntervalTime[i];
		avr += IntervalTime[i];
	}
	printf("Interval %10.3f: %10.3f : %10.3f\t",min,avr/(float)(TESTNUMBER-1),max);
        if (fabs(max-min)<=MAXJITTER)
        {
                printf("GOOD(fabs(%f)<%d micro sec)\n",max-min,MAXJITTER);
        }
        else
        {
                printf("FAIL(fabs(%f)>%d micro sec)\n",max-min,MAXJITTER);
        }


	return(0);
}


void *thread_main1(void *arg)
{
	static long err=-1;
	int class = 3;		/* FIFO Schedule */
	int priority = 2;			/* RealTime Priority min */
	static struct sched_param param;	/* RealTime Priority min */
	cpuset_t	*cpumask;
	cpu_t cpu = 1;				/* boot PROCESSOR */

	if (pthread_getschedparam(pthread_self(),&class,&param))
	{
                fprintf(stderr, "thread %d Cannot get priority %s\n",
                pthread_self(),strerror(errno));
	}
	param.sched_priority = priority;
	if (pthread_setschedparam(pthread_self(),SCHED_FIFO,(const struct sched_param *)&param))
	{
                fprintf(stderr, "thread %d Cannot set priority %s\n",
                pthread_self(),strerror(errno));
	}
        if (pthread_sigmask(SIG_SETMASK,&diset,&oset)==(-1))
        {
                fprintf(stderr, "Cannot set sigprocmask - %s\n",
                strerror(errno));
		err = errno;
                pthread_exit((void*)&err);
        }
#ifdef CCURRT
	cpumask = cpuset_alloc();
	cpuset_init(cpumask);
	cpuset_set_cpu(cpumask,cpu,1);/* CPU set */
	if ( mpadvise(MPA_PRC_SETBIAS,MPA_TID,0,cpumask)<0)
	{
                fprintf(stderr, "thread %d Cannot bind processor %s\n",
                pthread_self(),strerror(errno));
	}
        if (mpadvise (MPA_PRC_GETRUN,MPA_TID,0,cpumask)!=MPA_FAILURE)
        {
		char * cpustr;
                cpustr = cpuset_get_string(cpumask);
                printf("Thread %d is \"%s\" running\n",*(int*)arg,cpustr);
                free(cpustr);
        }
	cpuset_free(cpumask);
#endif
	MUTEX_LOCK(&Mutex[0]);
	Event[0]++;
	MUTEX_UNLOCK(&Mutex[0]);
        do
        {
		MUTEX_LOCK(&Mutex[1]); /* lock */
			if(Event[1]==0) COND_WAIT(&Condition[1],&Mutex[1]);
			readtime(&EndTime); 
			if (GlobalCounter<TESTNUMBER) adjust(&StartTime,&EndTime,&SwitchTime1[GlobalCounter/2]);
        		Event[1]--;
#if 0
                	printf("\t\tthread_main1 %d\n",Event[1]);
#endif
		MUTEX_UNLOCK(&Mutex[1]); /* unlock */
		/* event1_exec(); */
        } while (loop_flag);



	err=0;
        pthread_exit((void *)err);
	return(0);
}
void *thread_main2(void *arg)
{
	static long err=-1;
	int class = 3;				/* FIFO Schedule */
	int priority = 3;			/* RealTime Priority min */
	struct sched_param param;	/* RealTime Priority min */
	cpuset_t	*cpumask;
	cpu_t cpu = 1;				/* 2nd PROCESSOR */

	if (pthread_getschedparam(pthread_self(),&class,&param))
	{
                fprintf(stderr, "thread %d Cannot get priority %s\n",
                pthread_self(),strerror(errno));
	}
	param.sched_priority = priority;
	if (pthread_setschedparam(pthread_self(),SCHED_FIFO,(const struct sched_param *)&param))
	{
                fprintf(stderr, "thread %d Cannot set priority %s\n",
                pthread_self(),strerror(errno));
	}
        if (pthread_sigmask(SIG_SETMASK,&diset,&oset)==(-1))
        {
                fprintf(stderr, "Cannot set sigprocmask - %s\n",
                strerror(errno));
		err = errno;
                pthread_exit((void*)&err);
        }
#ifdef CCURRT
	cpumask = cpuset_alloc();
	cpuset_init(cpumask);
	cpuset_set_cpu(cpumask,cpu,1);/* CPU set */
	if ( mpadvise(MPA_PRC_SETBIAS,MPA_TID,0,cpumask)<0)
	{
                fprintf(stderr, "thread %d Cannot bind processor %s\n",
                pthread_self(),strerror(errno));
	}
        if (mpadvise (MPA_PRC_GETRUN,MPA_TID,0,cpumask)!=MPA_FAILURE)
        {
		char * cpustr;
                cpustr = cpuset_get_string(cpumask);
                printf("Thread %d is \"%s\" running\n",*(int*)arg,cpustr);
                free(cpustr);
        }
	cpuset_free(cpumask);
#endif
        MUTEX_LOCK(&Mutex[0]);
        Event[0]++;
        MUTEX_UNLOCK(&Mutex[0]);
        do
        {
                MUTEX_LOCK(&Mutex[2]); /* lock */
                        if (Event[2]==0) COND_WAIT(&Condition[2],&Mutex[2]);
                        readtime(&EndTime);
                        if (GlobalCounter<TESTNUMBER) adjust(&StartTime,&EndTime,&SwitchTime2[GlobalCounter/2]);
                        Event[2]--;
#if 0
                        printf("\t\t\t\tthread_main2 %d\n",Event[2]);
#endif
                MUTEX_UNLOCK(&Mutex[2]); /* unlock */
                /* event2_exec(); */
        } while (loop_flag);


       pthread_exit((void *)err);
	return(0);
}

void    *(*thread_main[THREAD_NO])(void *arg) =
{
	thread_main0,
	thread_main1,
	thread_main2
};



do_main(progname,fdin,filename,opt)
        char    *progname;
        FILE    *fdin;
        char    *filename;
	int opt;
{
	static int	no,thread_status[THREAD_NO],thread_arg[THREAD_NO];
	static pthread_t	tid[THREAD_NO];
	static pthread_attr_t	attributes[THREAD_NO];
	void *status_p;
	int signum;
        static struct sigaction newact;
        static sigset_t set,oset;
	static timer_t timer_id;

printf("*************************************************** \n");
printf("**** posix timer & thread contexst switch test **** \n");
printf("*************************************************** \n");


	if (opt)
	{
        	system("/usr/bin/shield -a 1");
        	system("/usr/bin/shield -c");
	}

        if (sigprocmask(0,NULL,&set)==(-1))
        {
                fprintf(stderr, "%s: Cannot get sigprocmask - %s\n",
		progname,strerror(errno));
		exit(1);
        }
        sigaddset(&set,SIGRTMIN+2);
        if (sigprocmask(SIG_SETMASK,&set,&oset)==(-1))
        {
                fprintf(stderr, "%s: Cannot set sigprocmask - %s\n",
		progname,strerror(errno));
		exit(1);
        }
	signum = SIGRTMIN+2;
	printf("Create POSIX timer %d Hz\n",CYCLE);
	timer_id = create_posix_timer(signum,0,1000000000/CYCLE);
	if (timer_id<0)
	{
               	fprintf(stderr, "%s: Cannot set posix timer - %s\n",
		progname,strerror(errno));
		exit(1);
	}
        /* Set up the signal handler using sigaction(2) or sigset(2) */
        sigemptyset(&newact.sa_mask);
        sigaddset(&newact.sa_mask,signum);
        newact.sa_handler = interval_hadler;
        newact.sa_flags = SA_SIGINFO|SA_RESTART;
        if (sigaction(signum, &newact, 0)==(-1))
        {
                fprintf(stderr, "%s: Cannot set sigaction - %s\n",
		progname,strerror(errno));
		exit(1);
        }

        sigfillset(&diset);
 	sigfillset(&eiset);
	sigdelset(&eiset,SIGRTMIN+2);
	sigdelset(&eiset,SIGALRM);
	sigdelset(&eiset,SIGINT);
#if 0
        do
        {
		sigsuspend(&eiset);
		if(GlobalCounter>=TESTNUMBER) loop_flag=0;
        } while (loop_flag);
#endif

	printf("Create %d thread\n",THREAD_NO);
	for (no=0;no<THREAD_NO;no++)
	{
		if (pthread_cond_init(&Condition[no],0)<0)
		{
                	fprintf(stderr, "%s: Cannot create condition - %s\n",
			progname,strerror(errno));
			exit(1);
		}
		if (pthread_mutex_init(&Mutex[no],0)<0)
		{
                	fprintf(stderr, "%s: Cannot create condition - %s\n",
			progname,strerror(errno));
			exit(1);
		}
		Event[no]=0;
	}
	for (no=0;no<THREAD_NO;no++)
	{
            	if (pthread_attr_init(&attributes[no]))
	    	{
                	fprintf(stderr, "%s: Cannot set attributes - %s\n",
			progname,strerror(errno));
			exit(1);
	    	}
                if (pthread_attr_setscope(&attributes[no],PTHREAD_SCOPE_SYSTEM))
	    	{
                	fprintf(stderr, "%s: Cannot set attributes - %s\n",
			progname,strerror(errno));
			exit(1);
	    	}
                if (pthread_attr_setdetachstate(&attributes[no],PTHREAD_CREATE_JOINABLE))
	    	{
                	fprintf(stderr, "%s: Cannot set attributes - %s\n",
			progname,strerror(errno));
			exit(1);
            	}
                if (pthread_attr_setinheritsched(&attributes[no],PTHREAD_INHERIT_SCHED))
	    	{
                	fprintf(stderr, "%s: Cannot set attributes - %s\n",
			progname,strerror(errno));
			exit(1);
            	}
	}
	MUTEX_LOCK(&Mutex[0]); /* lock */
	for (no=1;no<THREAD_NO;no++)
	{
		thread_arg[no] = no;
                thread_status[no] = pthread_create(&tid[no],&attributes[no],thread_main[no],(void *)&thread_arg[no]);
		if (thread_status[no])
		{
                	fprintf(stderr, "%s: Cannot create thread - %s\n",
			progname,strerror(errno));
			exit(1);
		}
	}
	MUTEX_UNLOCK(&Mutex[0]); /* unlock */
	thread_arg[0] = 0;
	thread_main[0]((void *)&thread_arg[0]);
	timer_delete(timer_id);
	for (no=1;no<THREAD_NO;no++)
	{
		pthread_join(tid[no],&status_p);
	}
}


main(argc, argv)
	int argc; 
	char **argv; 
{
    	extern int errno;
	extern int optind;
    	extern char *optarg;
    	register int i,c, opt = 0;


    	progname = argv[0];

    	while((c = getopt(argc, argv, "b")) != EOF){
		switch(c) {
			case 'b':
				opt =1;
				break;
	    		default:
				fprintf(stderr, 
				"usage: %s [-b]\n", 
				progname);
				exit(1);
    		}
	}
#if 0
    	if (optind != argc) {
		for (c = optind; c < argc; c++) {
    			if (access(argv[c], 4)) {
                               	fprintf(stderr, 
                                "%s: \"%s\" %s\n",
				 progname,argv[c],strerror(errno));
				exit(1);
    			}
		}
    	}
#endif
    	if (optind == argc) {
		do_main(progname,stdin, "(stdin)",opt);
    	} else {
		register FILE *fd;
		for (c = optind; c < argc; c++) {
    			if ((fd = fopen(argv[c], "w")) == NULL) {
                               	fprintf(stderr, 
                                "%s: \"%s\" %s\n",
				 progname,argv[c],strerror(errno));
				exit(1);
    			}
    			do_main(progname,fd,argv[c],opt);
			fprintf(fd,"Interval,Switch1,Switch2\n");
			for(i=0;i<TESTNUMBER/2;i++)
			{
				fprintf(fd,"%f,%f,%f\n",IntervalTime[i+1], SwitchTime1[i], SwitchTime2[i]);
			}
			for(i=TESTNUMBER/2;i<TESTNUMBER-1;i++)
			{
				fprintf(fd,"%f,,\n",IntervalTime[i+1]);
			}
    			fclose(fd);
		}
    	}
}
